--[[============================================================================
main.lua
============================================================================]]--

-- Variables & Globals

-- Table to hold global values
local nibbles = table.create()
function get_nibbles()
    --Pass globals to external files with this function
    return nibbles
end

nibbles.matrix_width = 80
nibbles.matrix_height = 50
nibbles.matrix_cells = table.create{}
nibbles.matrix_view = nil
nibbles.matrix_colors = table.create{'black','blue','yellow','green'} --Bitmaps dir

nibbles.color_bg = "black"
nibbles.color_snake = "blue"
nibbles.color_food = "yellow"
nibbles.color_wall = "green"

nibbles.current_dialog = nil
nibbles.score = 0
nibbles.snake = table.create()
nibbles.food = { x= 6, y= 6 }
nibbles.current_direction = "up"
nibbles.last_idle_time = os.clock()
nibbles.update_frequency = 0.08 --GUI refresh in seconds

nibbles.grow_amount = (1 * 4) -- This gets calculated again dynamically
nibbles.level_score = 0
nibbles.level_score_goal = 9
nibbles.level_defaults = table.create{}
--------------------------------------------------------------------------------
-- The Game
--------------------------------------------------------------------------------

-- Initialize high score

local prefs = renoise.Document.create("ScriptingToolPreferences") {
    high_score = 0
}
renoise.tool().preferences = prefs


-- Initialize levels
require "qbasic_levels"

-- Setup level rotation
local levels = {
    qbasic_level_1,
    qbasic_level_2,
    qbasic_level_3,
    qbasic_level_4,
    qbasic_level_5,
    qbasic_level_6,
    qbasic_level_7,
    qbasic_level_8,
    qbasic_level_9,
    qbasic_level_10,
}

local current_level = 1

--------------------------------------------------------------------------------

-- Reset Game

function reset()
    math.randomseed(os.time())

    nibbles.score = 0
    nibbles.level_score = 0

    current_level = 1
    levels[current_level]()

    draw_snake()
end

--------------------------------------------------------------------------------

-- Reset snake and place it

function snake_start(start_x, start_y, start_direction, start_length)
    -- Default start length at map/level start
    if not start_length then
        start_length = 3
    end
    -- Create snake table
    nibbles.snake = table.create()
    for i = 1, start_length do
        nibbles.snake[i] = { x = start_x, y = start_y }
    end
    -- Set direction
    nibbles.current_direction = start_direction
end

--------------------------------------------------------------------------------

-- Clear all matrix cells

function clear_cells()
    for x = 1,#nibbles.matrix_cells do
        for y = 1,#nibbles.matrix_cells[x] do
            clear_cell(x, y)
        end
    end
end

--------------------------------------------------------------------------------

-- Create main matrix cells

function create_cells()
    local vb = renoise.ViewBuilder()
    nibbles.matrix_view = vb:row { }

    for x = 1, nibbles.matrix_width do
        local column = vb:column { }
        nibbles.matrix_cells[x] = table.create()

        for y = 1, nibbles.matrix_height do
            nibbles.matrix_cells[x][y] = vb:bitmap {
                bitmap = "Bitmaps/cell_" .. nibbles.color_bg .. ".bmp",
            }
            column:add_child(nibbles.matrix_cells[x][y])
        end

        nibbles.matrix_view:add_child(column)
    end
end


--------------------------------------------------------------------------------

-- Draw snake

function draw_snake()
    for _,point in pairs(nibbles.snake) do
        set_cell_color(point.x, point.y, nibbles.color_snake)
    end
end


--------------------------------------------------------------------------------

-- Move snake

function move_snake()
    set_cell_color(nibbles.snake[1].x, nibbles.snake[1].y, nibbles.color_snake)
    set_cell_color(nibbles.snake[2].x, nibbles.snake[2].y, nibbles.color_snake)
end


--------------------------------------------------------------------------------

-- Access a cell in the matrix view

function matrix_cell(x, y)
    if (nibbles.matrix_cells[x] ~= nil) then
        return nibbles.matrix_cells[x][y]
    else
        return nil
    end
end


--------------------------------------------------------------------------------

-- Get the color of a cell

function get_cell_color(x ,y)
    local cell = matrix_cell(x, y)

    if (cell ~= nil) then
        local pos = cell.bitmap:find("Bitmaps/cell_")
        local color = cell.bitmap:sub(pos + 13)

        pos = color:find(".bmp")
        color = color:sub(1, pos - 1)

        return color
    end
end


--------------------------------------------------------------------------------

-- Set a cells color

function set_cell_color(x, y, color)
    assert(nibbles.matrix_colors:find(color), "invalid color")

    local cell = matrix_cell(x, y)
    if (cell ~= nil) then
        nibbles.matrix_cells[x][y].bitmap = "Bitmaps/cell_" .. color .. ".bmp"
    end
end


--------------------------------------------------------------------------------

-- Clear a cell

function clear_cell(x, y)
    set_cell_color(x, y, nibbles.color_bg)
end


--------------------------------------------------------------------------------

-- Keyboard input

function key_handler(dialog, key)

    if (key.name == "esc") then
        dialog:close()

    elseif (
        (key.name == "up" and nibbles.current_direction ~= "down") or
        (key.name == "down" and nibbles.current_direction ~= "up") or
        (key.name == "left" and nibbles.current_direction ~= "right") or
        (key.name == "right" and nibbles.current_direction ~= "left")
        )
        then
            nibbles.current_direction = key.name
        end

    end


    --------------------------------------------------------------------------------

    -- Start running the game logic (frame timer)

    function run()
        if not (renoise.tool().app_idle_observable:has_notifier(game)) then
            renoise.tool().app_idle_observable:add_notifier(game)
        end
    end


    --------------------------------------------------------------------------------

    -- Stop running the game (frame timer)

    function stop()
        if (renoise.tool().app_idle_observable:has_notifier(game)) then
            renoise.tool().app_idle_observable:remove_notifier(game)
        end
    end


    --------------------------------------------------------------------------------

    -- Game logic (frame timer)

    function game()

        -- Game was closed?
        if (not nibbles.current_dialog or not nibbles.current_dialog.visible) then
            stop()
            return
        end

        -- Only run every 0.1 seconds
        if (os.clock() - nibbles.last_idle_time < nibbles.update_frequency) then
            return
        end

        -- Draw food
        set_cell_color(nibbles.food.x, nibbles.food.y, nibbles.color_food)

        -- Do frame stuff
        local tmp1 = table.create{x=nil, y=nil} --Init with empty/useless coordinates
        local tmp2 = table.create{x=nil, y=nil} --Ditto

        tmp1.x = nibbles.snake[1].x
        tmp1.y = nibbles.snake[1].y

        -- Check the direction
        if (nibbles.current_direction == "up") then
            nibbles.snake[1].y = nibbles.snake[1].y - 1
        elseif (nibbles.current_direction == "down") then
            nibbles.snake[1].y = nibbles.snake[1].y + 1
        elseif (nibbles.current_direction == "left") then
            nibbles.snake[1].x = nibbles.snake[1].x - 1
        elseif (nibbles.current_direction == "right") then
            nibbles.snake[1].x = nibbles.snake[1].x + 1
        else
            return
        end

        -- Snake crashed
        if (
            nibbles.snake[1].x < 1 or
            nibbles.snake[1].x > nibbles.matrix_width or
            nibbles.snake[1].y < 1 or
            nibbles.snake[1].y > nibbles.matrix_height or
            nibbles.color_snake == get_cell_color(nibbles.snake[1].x, nibbles.snake[1].y) or
            nibbles.color_wall == get_cell_color(nibbles.snake[1].x, nibbles.snake[1].y)
            ) then
            if nibbles.score > prefs.high_score.value then
                prefs.high_score.value = nibbles.score
            end
            renoise.app():show_error(
            "Game over!" .. "\nYour score is: " .. nibbles.score ..
            "\nYour high score is: " .. prefs.high_score.value
            )
            nibbles.current_dialog:show()
            reset()

            return
        end

        -- Snake ate some Food, so he grows
        if (nibbles.snake[1].x == nibbles.food.x and nibbles.snake[1].y == nibbles.food.y) then
            -- Grow
            clear_cell(nibbles.food.x, nibbles.food.y)
            for _ = 1, nibbles.grow_amount do
                nibbles.snake[#nibbles.snake + 1] = table.create{x=nil, y=nil}
                nibbles.snake[#nibbles.snake].x = nibbles.snake[#nibbles.snake-1].x
                nibbles.snake[#nibbles.snake].y = nibbles.snake[#nibbles.snake-1].y
            end
            -- Update grow amount for next number
            nibbles.grow_amount = (nibbles.level_score + 1) * 4
            -- Increment score
            nibbles.score = nibbles.score + 1
            nibbles.level_score = nibbles.level_score + 1
            -- Display score
            renoise.app():show_status("Nibbles:  "..nibbles.level_score.."/"..nibbles.level_score_goal.."   ("..nibbles.score..")")
            -- Next level?
            if nibbles.level_score == nibbles.level_score_goal then
                renoise.app():show_message(
                "Level finished!" .. "\nYour score is: " .. nibbles.score ..
                "\nYour high score is: " .. prefs.high_score.value
                )
                nibbles.current_dialog:show()
                current_level = (current_level < #levels and current_level + 1) or 1
                levels[current_level]()
                nibbles.level_score = 0
                nibbles.grow_amount = (nibbles.level_score + 1) * 4
            end
            -- New food
            local new_food = table.create()
            repeat
                local fa = nibbles.level_defaults.food_area
                new_food.x = fa.x_min + math.random(fa.x_max - fa.x_min)
                new_food.y = fa.y_min + math.random(fa.y_max - fa.y_min)
            until get_cell_color(new_food.x, new_food.y) == nibbles.color_bg
            nibbles.food.x = new_food.x
            nibbles.food.y = new_food.y
        end

        -- Move the Snake
        local last_block = nibbles.snake[#nibbles.snake]
        clear_cell(nibbles.snake[#nibbles.snake].x, nibbles.snake[#nibbles.snake].y)
        for i=2, table.count(nibbles.snake) do
            tmp2.x = nibbles.snake[i].x
            tmp2.y = nibbles.snake[i].y
            nibbles.snake[i].x = tmp1.x
            nibbles.snake[i].y = tmp1.y
            tmp1.x = tmp2.x
            tmp1.y = tmp2.y
        end

        -- Update cells
        move_snake()

        -- Memorize time for the frame timer
        nibbles.last_idle_time = os.clock()

    end


    --------------------------------------------------------------------------------

    -- Initializes and shows the game

    function create_game()

        if (not nibbles.current_dialog or not nibbles.current_dialog.visible) then
            create_cells()
            reset()
            run()

            nibbles.current_dialog = renoise.app():show_custom_dialog(
            "Nibbles", nibbles.matrix_view, key_handler)
        end

    end


    --------------------------------------------------------------------------------
    -- Menu Registration
    --------------------------------------------------------------------------------

    renoise.tool():add_menu_entry {
        name = "Main Menu:Tools:Nibbles...",
        invoke = create_game
    }

